#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <sys/stat.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>
#include "simtm.h"

char save_flg, log_flg, send_jis_flg;
unsigned char ch_db_buf, connect_mes[10];
int siofd, pid, pipe_fd[2], logfd;
int connect_flg, connect_cnt, ch_db_flg;
int default_kanji_code, kanji_code;
Widget text;
fd_set  r_mask;
extern struct termios sio_buf;

static XtActionsRec actionTable[] =
{  {"send",Send},
};

static char defaultTranslations[] = "<Key>Return: send()";

void Int_select()
{
  FD_ZERO(&r_mask);
  FD_SET(siofd, &r_mask);
  FD_SET(pipe_fd[0], &r_mask);
}

void Save_log_file()
{
  struct tm *tm_ptr;
  time_t the_time;
  unsigned char log_name[8];

  if( log_flg == 0 )
  {
    time(&the_time);
    tm_ptr = localtime(&the_time);
    sprintf(log_name, "%02d%02d.log", tm_ptr->tm_mon+1, tm_ptr->tm_mday);
    logfd = open(log_name, O_WRONLY | O_APPEND | O_CREAT, S_IRUSR | S_IWUSR);
    printf("\n**********  Save Log File (%s)  **********\n",log_name);
    log_flg = 1;
  }
}

void Close_log_file()
{
  if( log_flg == 1 )
  {
    close(logfd);
    printf("\n**********  Close Log File  **********\n");
    log_flg = 0;
    connect_flg = 0;
  }
}

void Read_pipe()
{
  unsigned char read_mes;

  read(pipe_fd[0], &read_mes, 1);

  if( read_mes == 0x00 && log_flg == 1 )
    Close_log_file();

  if( read_mes & 0x01 && log_flg == 0 )
    Save_log_file();

  if( read_mes & 0x02 )
    connect_flg = 1;

  if( read_mes & 0x40 )
    pause();

  if( read_mes & 0x80 )
    exit(0);

  return;
}

void Auto_save( ch_in )  /*  auto save logfile  */
unsigned char ch_in;
{
  if( connect_cnt < 11 )
  {
    connect_mes[connect_cnt] = ch_in;
    connect_cnt += 1;
  }
  if( ch_in == 0x0d || ch_in == 0x0a )
  {
    if( strncmp(connect_mes, "CONNECT", 7) == 0 )
      Save_log_file();

    if( strncmp(connect_mes, "NO CARRIER", 10) == 0 )
      Close_log_file();

    memset( connect_mes, 0, sizeof(connect_mes) );
    connect_cnt = 0;
    return;
  }
}

void Read_data()  /*  read data from RS-232C  */
{
  unsigned char ch_in, buf[1024];
  int i, n;

  n = read(siofd, &buf, 1024);

  for( i = 0; i < n; i ++ )
  {
    ch_in = buf[i];
    if( kanji_code == 0 )
      Res_JIS_code( ch_in );

    else if( kanji_code == 1 )
      Res_EUC_code( ch_in );

    else if( kanji_code == 2 )
      Res_SJIS_code( ch_in );
  }

  return;
}

void print_mes(sig)
int sig;
{
}

void Receve()  /*  receive data or pipe  */
{
  fd_set  readOk;

  Int_select();

  log_flg = 0;

  signal( SIGCONT , print_mes );

  while(1)
  {
    readOk = r_mask;
    select(FD_SETSIZE, &readOk, NULL, NULL, NULL);

    if ( FD_ISSET(siofd, &readOk) )
      Read_data();     /* receive from RS-232C */
    if ( FD_ISSET(pipe_fd[0], &readOk) )
      Read_pipe();     /* receive parent process */
  }
}

void Send_text( str, ch_len )  /* send text */
unsigned char *str;
int ch_len;
{
  int i;
  unsigned char db1, db2, ch_cr, *ch_db;
  unsigned char st_jis[3] = {0x1b,0x24,0x42},
	        st_roma[3] = {0x1b,0x28,0x4a};

  i = 0;

  while( i < ch_len )
  {
    if( str[i] & 0x80 )
    {
      if( kanji_code == 0 )   /* outputcode = JIS */
      {
	if( str[i] == 0x8e )  /*  if text code is hankaku-kana  */
	{
	  if( send_jis_flg == 1 )
	  {
	    write(siofd, st_roma, 3);  /* output start code (romaji)  */
	    send_jis_flg = 0;
	  }
	  write(siofd, &str[i+1], 1);
	}
	else
	{
	  db1 = str[i] & 0x7f;
	  db2 = str[i+1] & 0x7f;
	  if( send_jis_flg == 0 )
	    write(siofd, st_jis, 3);  /* output start code (JIS)  */
	  write(siofd, &db1, 1);
	  write(siofd, &db2, 1);
	  send_jis_flg = 1;
	}
      }
      if( kanji_code == 1 )  /* outputcode =  EUC */
      {
	db1 = str[i];
	db2 = str[i+1];
	write(siofd, &db1, 1);
	write(siofd, &db2, 1);
      }
      if( kanji_code == 2 )  /* outputcode = S-JIS */
      {
	if( str[i] == 0x8e )  /*  if text code is hankaku-kana  */
	  write(siofd, &str[i+1], 1);
	else
	{
	  ch_db = Euc2sjis( str[i], str[i+1] );  /* change EUC to S-JIS */
	  write(siofd, ch_db, 2);
	}
      }
      i += 2;
    }
    else
    {
      if( kanji_code == 0 && send_jis_flg == 1 )
      {
	write(siofd, st_roma, 3);  /* output start code (romaji)  */
	send_jis_flg = 0;
      }
      write(siofd, &str[i], 1);
      i++;
    }
  }
  if( kanji_code == 0 && send_jis_flg == 1 )
  {
    write(siofd, st_roma, 3);  /* output start code (romaji)  */
    send_jis_flg = 0;
  }

  ch_cr = 0x0d;

  write(siofd, &ch_cr, 1);
}

void Send()   /* send TextWidget to RS-232C */
{
  String str;
  int ch_len;

  XtVaGetValues(text, XtNstring, &str, NULL);
  ch_len = strlen(str);

  Send_text( str, ch_len );

  XtVaSetValues(text, XtNstring, "", NULL);
}

void Save(w)  /* write Logfile */
Widget w;
{
  if( !(save_flg & 0x01) )
  {
    save_flg = 1;
    write(pipe_fd[1], &save_flg, 1);
  }

}

void Close(w)  /* close Logfile */
Widget w;
{
  if( save_flg & 0x01 )
  {
    save_flg = 0x00;
    write(pipe_fd[1], &save_flg, 1);
  }
}

void Quit()
{
  char f_flg;

  if(tcsetattr(siofd, TCSANOW, &sio_buf) == -1)
  {
    printf("tcsetattr error\n");
  };

  close(siofd);

  f_flg = 0x00;
  write(pipe_fd[1], &f_flg, 1);    /* close Logfile */
  f_flg = 0x80;
  write(pipe_fd[1], &f_flg, 1);    /* end child process */

  exit(0);
}

void Key_event( w, client, key_event )  /* input except TextWidget */
Widget w;
caddr_t  client;
XEvent *key_event;
{
  char key_ch;
  KeySym  key;

  if( key_event->type == KeyPress )
  {
    XLookupString( key_event, &key_ch, 1, &key, NULL );
    switch(key)
    {
      case XK_Shift_R :
	break;
      case XK_Shift_L :
	break;
      default :
        write(siofd, &key_ch, 1);
    }
  }
}

void Delete( w, event, params, num_params )  /*  Delete main  */
Widget w;
XEvent *event;
String *params;
int    num_params;
{
  Quit();
}

void main(int argc, char **argv)
{
  Widget top, form, dial, save_log, close_log, text_up, shell, hangup, quit;
  XtTranslations trans_table;
  static XtActionsRec actions[] = {{"delete", Delete}};
  static Atom wm_delete_window;

  default_kanji_code = 2;  /*  default is SJIS code  */

  Read_ini_file();

  kanji_code = default_kanji_code;
  send_jis_flg = 0;


  if( pipe(pipe_fd) == -1 )
  {
    perror("pipe");
    exit(1);
  }

  save_flg = 0;
  connect_flg = 0;
  connect_cnt = 0;
  ch_db_flg = 0;

  memset( connect_mes, 0, sizeof(connect_mes) );

  XtSetLanguageProc(NULL, NULL, NULL);
  top = XtInitialize(argv[0], "Simtmrc", NULL, 0, &argc, argv);

  form   = XtVaCreateManagedWidget("form", formWidgetClass, top, NULL);
  text   = XtVaCreateManagedWidget("text", asciiTextWidgetClass, form,
				   XtNeditType, XawtextEdit,
				   NULL);
  dial = XtVaCreateManagedWidget("dial", commandWidgetClass, form,
				 XtNlabel, (XtArgVal)"Dial",
				 NULL);
  save_log = XtVaCreateManagedWidget("save", commandWidgetClass, form,
				     XtNlabel, (XtArgVal)"Save log",
				     NULL);
  close_log = XtVaCreateManagedWidget("close", commandWidgetClass, form,
				      XtNlabel, (XtArgVal)"Close log",
				      NULL);
  text_up = XtVaCreateManagedWidget("fileup", commandWidgetClass, form,
				    XtNlabel, (XtArgVal)"Text Upload",
				    NULL);
  shell = XtVaCreateManagedWidget("shell_com", commandWidgetClass, form, NULL);
  hangup = XtVaCreateManagedWidget("hangup", commandWidgetClass, form, NULL);
  quit = XtVaCreateManagedWidget("quit", commandWidgetClass, form, NULL);

  Dialpopup(dial);
  Filepopup(text_up);
  Shellpopup(shell);

  XtAddActions(actionTable,XtNumber(actionTable));
  trans_table = XtParseTranslationTable(defaultTranslations);
  XtOverrideTranslations(text, trans_table);

  XtAddCallback(dial, XtNcallback, Open_dial_popup, NULL);
  XtAddCallback(save_log, XtNcallback, Save, NULL);
  XtAddCallback(close_log, XtNcallback, Close, NULL);
  XtAddCallback(text_up, XtNcallback, Open_file_popup, NULL);
  XtAddCallback(shell, XtNcallback, Open_shell_popup, NULL);
  XtAddCallback(hangup, XtNcallback, Hangup, NULL);
  XtAddCallback(quit, XtNcallback, Quit, NULL);

  XtAddEventHandler( form, KeyPressMask, False, Key_event, NULL );
          /* add EventHandler to FormWidget */

  XtAddActions( actions, XtNumber(actions) );
  XtOverrideTranslations( top,
	      XtParseTranslationTable("<Message>WM_PROTOCOLS: delete()"));
  wm_delete_window = XInternAtom(XtDisplay(top), "WM_DELETE_WINDOW", False);

  XtRealizeWidget(top);

  XSetWMProtocols( XtDisplay(top), XtWindow(top), &wm_delete_window, 1 );

  if( (pid = fork()) == 0 ) /* run child process */
  {
    close(pipe_fd[1]);
    Receve();
  }
  if( pid >= 1 )
  {
    close(pipe_fd[0]);
    while(1)
    {
      XEvent    event;

      XtNextEvent(&event);
      XtDispatchEvent(&event);
     }
  }
}
